@@ -65,6 +65,12 @@ def parse_args(msg):
         default=False,
         help="Set to disable TLS encryption on the gRPC channel",
     )
+    parser.add_argument(
+        "--skip-verify",
+        action="store_true",
+        default=False,
+        help="Set to disable SSL certificate valication on the encrypted gRPC channel",
+    )
     parser.add_argument(
         "-o", "--operation",
         type=str,
@@ -83,6 +89,13 @@ def parse_args(msg):
         default="", nargs="+",
         help="gNMI paths of interest in XPath format, space separated"
     )
+    parser.add_argument(
+        "--gnmi_path_target",
+        type=str,
+        required=False,
+        default="",
+        help="Set target for GNMI path if it is different to the endpoint itself."
+    )
     parser.add_argument(
         "-d", "--datastore",
         type=str,
